{-------------------------------------------------------------------------------

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

     The Original Code is ButtonBars.pas released at May 26th, 2007.
     The Initial Developer of the Original Code is Priyatna.
     (Website: http://www.priyatna.org Email: me@priyatna.org)
     All Rights Reserved.

Alternatively, the contents of this file may be used under the terms of the
GNU General Public License Version 2 or later (the "GPL"), in which case
the provisions of the GPL are applicable instead of those above.
If you wish to allow use of your version of this file only under the terms
of the GPL and not to allow others to use your version of this file
under the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the GPL.
If you do not delete the provisions above, a recipient may use your version
of this file under either the MPL or the GPL.

-------------------------------------------------------------------------------}

unit ButtonBars;

{$MODE Delphi}

interface

uses
  Classes, Graphics, Controls, ExtCtrls, SysUtils, ImgList, Messages,
  Forms, Themes;

type
  TBarDirection = (bdHorizontal, bdVertical);

  TButtonBar = class(TCustomControl)
  private
    FShowCaptions: Boolean;
    FButtonHeight: Integer;
    FItemIndex: Integer;
    FButtonWidth: Integer;
    FDirection: TBarDirection;
    FImages: TCustomImageList;
    FOnChange: TNotifyEvent;
    FHints: TStringList;
    FButtons: TStringList;
    FChangeLink: TChangeLink;
    FAutoSize: Boolean;
    LastHover: Integer;
    procedure SetButtonHeight(const Value: Integer);
    procedure SetButtons(const Value: TStringList);
    procedure SetButtonWidth(const Value: Integer);
    procedure SetDirection(const Value: TBarDirection);
    procedure SetHints(const Value: TStringList);
    procedure SetImages(const Value: TCustomImageList);
    procedure SetItemIndex(const Value: Integer);
    procedure SetShowCaptions(const Value: Boolean);
    procedure ButtonsChange(Sender: TObject);
    function GetRect(Index: Integer): TRect;
    procedure DrawItem(Index: Integer);
    procedure ChangeLinkChange(Sender: TObject);
    procedure SetAutoSize(const Value: Boolean); reintroduce;
    procedure AutoResizeControl;
    procedure WMSize(var Msg: TWMSize); message WM_SIZE;
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
    procedure Paint; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    function GetIndexFromXY(const X, Y: Integer): Integer;
  published
    property ButtonWidth: Integer read FButtonWidth write SetButtonWidth;
    property ButtonHeight: Integer read FButtonHeight write SetButtonHeight;
    property Images: TCustomImageList read FImages write SetImages;
    property Buttons: TStringList read FButtons write SetButtons;
    property Hints: TStringList read FHints write SetHints;
    property ItemIndex: Integer read FItemIndex write SetItemIndex;
    property ShowCaptions: Boolean read FShowCaptions write SetShowCaptions;
    property Direction: TBarDirection read FDirection write SetDirection;
    property AutoSize: Boolean read FAutoSize write SetAutoSize;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
    property Align;
    property Anchors;
    property ShowHint;
  end;

  procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('priyatna.org', [TButtonBar]);
end;


{ TButtonBar }

constructor TButtonBar.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FButtonWidth := 80;
  FButtonHeight := 80;
  FItemIndex := -1;
  FShowCaptions := True;
  FDirection := bdVertical;
  FButtons := TStringList.Create;
  FButtons.OnChange := ButtonsChange;
  FHints := TStringList.Create;
  Color := clBtnFace;
  LastHover := -1;
  FChangeLink := TChangeLink.Create;
  FChangeLink.OnChange := ChangeLinkChange;
end;

destructor TButtonBar.Destroy;
begin
  if Assigned(FImages)
    then FImages.UnRegisterChanges(FChangeLink);
  FChangeLink.Free;
  FButtons.Free;
  FHints.Free;
  inherited;
end;

procedure TButtonBar.Notification(AComponent: TComponent;
  Operation: TOperation);
begin
  if (AComponent = FImages) and (Operation = opRemove) then
  begin
    if Assigned(FImages)
      then FImages.UnRegisterChanges(FChangeLink);
    FImages := nil;
    Invalidate;
  end;
  inherited;
end;

procedure TButtonBar.MouseDown(Button: TMouseButton; Shift: TShiftState; X,
  Y: Integer);
var
  i: Integer;  
begin
  if (Button = mbLeft) and (Shift = [ssLeft]) then
  begin
    i := GetIndexFromXY(X, Y);
    if (i <> -1) and (FItemIndex <> i) then
    begin
      SetItemIndex(i);
      if Assigned(FOnChange)
        then FOnChange(Self);
    end;
  end;
  inherited;
end;

procedure TButtonBar.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  Index: Integer;
  Rc: TRect;
  Pt: TPoint;
begin
  if ShowHint then
  begin
    Index := GetIndexFromXY(X, Y);
    if (Index <> LastHover) then
    begin
      if (Index < FHints.Count) and (Index >= 0)
        then Hint := FHints[Index]
        else Hint := '';
      if (Index <> -1) and (Hint <> '') then
      begin
        Rc := GetRect(Index);
        Pt := ClientToScreen(Point(Rc.Left, Rc.Bottom));
        Application.CancelHint;
        Application.ActivateHint(Pt);
      end;
      LastHover := Index;
    end;    
  end;
end;

procedure TButtonBar.SetButtonHeight(const Value: Integer);
begin
  if (FButtonHeight <> Value) then
  begin
    FButtonHeight := Value;
    Invalidate;
  end;  
end;

procedure TButtonBar.SetButtonWidth(const Value: Integer);
begin
  if (FButtonWidth <> Value) then
  begin
    FButtonWidth := Value;
    Invalidate;
  end;
end;

procedure TButtonBar.SetDirection(const Value: TBarDirection);
begin
  if (FDirection <> Value) then
  begin
    FDirection := Value;
    Invalidate;
  end;
end;

procedure TButtonBar.SetButtons(const Value: TStringList);
begin
  FButtons.Assign(Value);
end;

procedure TButtonBar.ButtonsChange(Sender: TObject);
begin
  if (FButtons.Count = 0)
    then FItemIndex := -1;
  if (FItemIndex >= FButtons.Count)
    then FItemIndex := -1;
  AutoResizeControl;   
  Invalidate;
end;

procedure TButtonBar.SetHints(const Value: TStringList);
begin
  FHints.Assign(Value);
end;

procedure TButtonBar.SetImages(const Value: TCustomImageList);
begin
  if (FImages <> Value) then
  begin
    if Assigned(FImages)
      then FImages.UnRegisterChanges(FChangeLink);
    FImages := Value;
    if Assigned(FImages)
      then FImages.RegisterChanges(FChangeLink);
    Invalidate;
  end;
end;

procedure TButtonBar.ChangeLinkChange(Sender: TObject);
begin
  Invalidate;
end;

procedure TButtonBar.SetItemIndex(const Value: Integer);
var
  Old: Integer;
begin
  if (FItemIndex <> Value) then
  begin
    Old := FItemIndex;
    FItemIndex := Value;
    if (FButtons.Count = 0)
      then FItemIndex := -1;
    if (FItemIndex >= FButtons.Count)
      then FItemIndex := -1;
    if (Old <> -1)
      then DrawItem(Old);
    if (FItemIndex <> -1)
      then DrawItem(FItemIndex);  
  end;
end;

procedure TButtonBar.SetShowCaptions(const Value: Boolean);
begin
  if (FShowCaptions <> Value) then
  begin
    FShowCaptions := Value;
    Invalidate;
  end;
end;

function TButtonBar.GetRect(Index: Integer): TRect;
var
  l, t: Integer;
begin
  if (FDirection = bdHorizontal) then
  begin
    l := Index * FButtonWidth;
    t := 0;
  end else
  begin
    l := 0;
    t := Index * FButtonHeight;
  end;
  Result := Rect(l, t, l + FButtonWidth, t + FButtonHeight);
end;

function TButtonBar.GetIndexFromXY(const X, Y: Integer): Integer;
begin
  Result := -1;
  if (FDirection = bdHorizontal) then
  begin
    if (Y < FButtonHeight) then
    begin
      Result := (X div FButtonWidth) - 1;
      if ((X mod FButtonWidth) > 0)
        then Inc(Result);
    end;
  end else
  begin
    if (X < FButtonWidth) then
    begin
       Result := (Y div FButtonHeight) - 1;
       if ((Y mod FButtonHeight) > 0)
         then Inc(Result);
    end;
  end;
  if (FButtons.Count = 0) or (Result >= FButtons.Count)
    then Result := -1;
end;

procedure TButtonBar.Paint;
var
  x: Integer;
begin
  for x := 0 to Buttons.Count-1
    do DrawItem(x);
end;

procedure TButtonBar.DrawItem(Index: Integer);
var
  Rc: TRect;
  l, t, h, ih: Integer;
  //Details: TThemedElementDetails;
begin
  Rc := GetRect(Index);
  with Canvas do
  begin
    Brush.Style := bsSolid;

    if (Index = FItemIndex) then
    begin
      Brush.Color := clBtnShadow;
      FillRect(Rc);
      Canvas.Frame3D(Rc, 1, bvLowered);
    end else
    begin
      Brush.Color := clBtnFace;
      FillRect(Rc);
      Canvas.Frame3D(Rc, 1, bvRaised);
    end;
    {
    if ThemeServices.ThemesAvailable then
    begin
      Brush.Color := clBtnFace;
      FillRect(Rc);
      if (Index = FItemIndex)
        then Details := ThemeServices.GetElementDetails(ttbButtonPressed)
        else Details := ThemeServices.GetElementDetails(ttbButtonNormal);
      ThemeServices.DrawElement(Canvas.Handle, Details, Rc);
    end;
    }
    h := 0;
    ih := 0;
    if Assigned(FImages) then
    begin
      ih := FImages.Height;
      h := ih;
    end;
    if FShowCaptions
      then h := h + TextHeight('Ag');
    t := Rc.Top + ((FButtonHeight - h) div 2);
    if Assigned(FImages) then
    begin
      l := Rc.Left + ((FButtonWidth - FImages.Width) div 2);
      FImages.Draw(Canvas, l, t, Index);
    end;
    if FShowCaptions then
    begin
      t := t + ih;
      l := Rc.Left + ((FButtonWidth - TextWidth(FButtons[Index])) div 2);
      Brush.Style := bsClear;
      TextRect(Rc, l, t, FButtons[Index]);
    end;
  end;
end;

procedure TButtonBar.AutoResizeControl;
begin
  if FAutoSize and (FButtons.Count > 0) then
  begin
    if (FDirection = bdHorizontal) then
    begin
      Width := FButtons.Count * FButtonWidth;
      Height := FButtonHeight;
    end else
    begin
      Width := FButtonWidth;
      Height := FButtons.Count * FButtonHeight;
    end;
  end;
end;

procedure TButtonBar.SetAutoSize(const Value: Boolean);
begin
  if (FAutoSize <> Value) then
  begin
    FAutoSize := Value;
    AutoResizeControl;
  end;  
end;

procedure TButtonBar.WMSize(var Msg: TWMSize);
begin
  inherited;
  if FAutoSize
    then AutoResizeControl;
end;

end.

