{-------------------------------------------------------------------------------

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

     The Original Code is NiceGallery.pas released at May 26th, 2007.
     The Initial Developer of the Original Code is Priyatna.
     (Website: http://www.priyatna.org Email: me@priyatna.org)
     All Rights Reserved.
     
     Contributors:
     - Carl Stuffer (cstuffer@hotmail.com), fixing autoscroll bugs

Alternatively, the contents of this file may be used under the terms of the
GNU General Public License Version 2 or later (the "GPL"), in which case
the provisions of the GPL are applicable instead of those above.
If you wish to allow use of your version of this file only under the terms
of the GPL and not to allow others to use your version of this file
under the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the GPL.
If you do not delete the provisions above, a recipient may use your version
of this file under either the MPL or the GPL.

-------------------------------------------------------------------------------}

unit NiceGallery;

{$MODE Delphi}

interface

uses
  LCLIntf, LCLType, LMessages, Classes, Controls, Graphics, ExtCtrls, Forms;

type

  TNiceGallery = class;

  TGalleryItem = class(TCollectionItem)
  private
    DisplayRect: TRect;
    FTag: Integer;
    FBitmap: TBitmap;
    FHint: string;
    function GetGallery: TNiceGallery;
    procedure BitmapChange(Sender: TObject);
    procedure SetBitmap(Value: TBitmap);
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    property Gallery: TNiceGallery read GetGallery;
  published
    property Tag: Integer read FTag write FTag;
    property Hint: string read FHint write FHint;
    property Bitmap: TBitmap read FBitmap write SetBitmap;
  end;

  TGalleryItems = class(TCollection)
  private
    FGallery: TNiceGallery;
    function GetItem(Index: Integer): TGalleryItem;
    procedure SetItem(Index: Integer; Value: TGalleryItem);
  protected
    function GetOwner: TPersistent; override;
    procedure Update(Item: TCollectionItem); override;
  public
    constructor Create(AGallery: TNiceGallery);
    property Gallery: TNiceGallery read FGallery;
    property Items[Index: Integer]: TGalleryItem read GetItem write SetItem; default;
    function Add: TGalleryItem;
    function AddItem(Item: TGalleryItem; Index: Integer): TGalleryItem;
    function Insert(Index: Integer): TGalleryItem;
  end;

  TGalleryEvent = procedure (Sender: TObject; Index: Integer) of object;

  TNiceGallery = class(TCustomControl)
  private
    LScroller: TRect;
    RScroller: TRect;
    LArrow, RArrow: array of TPoint;
    LScrollerVisible: Boolean;
    RScrollerVisible: Boolean;
    Delta: Integer;
    MaxDelta: Integer;
    LastHover: Integer;
    LastScroll: Integer;
    HoverIndex: Integer;
    RgnL, RgnR: HRGN;
    MustScrollLeft, MustScrollRight: Boolean;
    AutoScroll: Boolean;
    CanScroll: Boolean;

    FScrollerSize: Integer;
    FScrollerArrowColor: TColor;
    FScrollerColor: TColor;
    FScrollerHoverArrowColor: TColor;
    FScrollerHoverColor: TColor;
    FHandPointCursor: Boolean;
    FSelectionColor: TColor;
    FItems: TGalleryItems;
    FTimer: TTimer;
    FOnHover: TGalleryEvent;
    FOnClick: TGalleryEvent;
    FLargeChange: Integer;
    FSmallChange: Integer;
    FActive: Boolean;

    function CreateItem: TGalleryItem;
    procedure UpdateItem(Index: Integer);
    procedure UpdateItems;
    procedure Recalculate;
    procedure CheckScroller;
    procedure DrawItem(Index: Integer);
    function GetItemAtXY(X, Y: Integer): Integer;
    procedure TimerTimer(Sender: TObject);

    procedure CMColorChanged(var Msg: TLMessage); message CM_COLORCHANGED;
    procedure CMMouseLeave(var Msg: TLMessage); message CM_MOUSELEAVE;
    procedure WMSize(var Msg: TLMSize); message LM_SIZE;
    procedure WMEraseBkgnd(var Msg: TLMessage); message LM_ERASEBKGND;
    procedure SetItems(Value: TGalleryItems);
    procedure SetScrollerSize(Value: Integer);
    procedure SetScrollerColor(Value: TColor);
    procedure SetScrollerArrowColor(Value: TColor);
    procedure SetActive(const Value: Boolean);

  protected
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;

  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

  published
     property Active: Boolean read FActive write SetActive default False;
     property Items: TGalleryItems read FItems write SetItems;
     property ScrollerSize: Integer read FScrollerSize write SetScrollerSize default 25;
     property ScrollerArrowColor: TColor read FScrollerArrowColor write SetScrollerArrowColor default clWhite;
     property ScrollerColor: TColor read FScrollerColor write SetScrollerColor default clBlack;
     property ScrollerHoverArrowColor: TColor read FScrollerHoverArrowColor write FScrollerHoverArrowColor default clYellow;
     property ScrollerHoverColor: TColor read FScrollerHoverColor write FScrollerHoverColor default clBlack;
     property HandPointCursor: Boolean read FHandPointCursor write FHandPointCursor default True;
     property SelectionColor: TColor read FSelectionColor write FSelectionColor default clRed;
     property SmallChange: Integer read FSmallChange write FSmallChange default 3;
     property LargeChange: Integer read FLargeChange write FLargeChange default 30;

     property OnHover: TGalleryEvent read FOnHover write FOnHover;
     property OnClick: TGalleryEvent read FOnClick write FOnClick;

     property Color;
     property TabStop;
     property TabOrder;
     property Anchors;
     property Align;
     property ShowHint;

  end;


implementation

uses
  Math;

const
  SCLEFTINDEX  = 2147483647;
  SCRIGHTINDEX = 2147483646;


{ TGalleryItem }

constructor TGalleryItem.Create(Collection: TCollection);
begin
  FBitmap := TBitmap.Create;
  FBitmap.OnChange := BitmapChange;
  inherited Create(Collection);
end;

destructor TGalleryItem.Destroy;
begin
  inherited Destroy;
  FBitmap.Free;
end;

function TGalleryItem.GetGallery: TNiceGallery;
begin
  Result := TGalleryItems(Collection).FGallery;
end;

procedure TGalleryItem.Assign(Source: TPersistent);
begin
  if (Source is TGalleryItem) then
  begin
    FTag := TGalleryItem(Source).Tag;
    FHint := TGalleryItem(Source).Hint;
    FBitmap.Assign(TGalleryItem(Source).Bitmap);
    //Changed(True);
  end;
end;

procedure TGalleryItem.SetBitmap(Value: TBitmap);
begin
  FBitmap.Assign(Value);
end;

procedure TGalleryItem.BitmapChange(Sender: TObject);
begin
  Changed(True);
end;


{ TGalleryItems }

constructor TGalleryItems.Create(AGallery: TNiceGallery);
begin
  inherited Create(TGalleryItem);
  FGallery := AGallery;
end;

function TGalleryItems.Add: TGalleryItem;
begin
  Result := TGalleryItem(inherited Add);
end;

function TGalleryItems.AddItem(Item: TGalleryItem;
  Index: Integer): TGalleryItem;
begin
  if (Item = nil)
    then Result := FGallery.CreateItem
    else
    begin
      Result := Item;
      if Assigned(Item) then
      begin
        Result.Collection := Self;
        if (Index < 0)
          then Index := Count - 1;
        Result.Index := Index;
      end;
    end;
end;

function TGalleryItems.GetItem(Index: Integer): TGalleryItem;
begin
  Result := TGalleryItem(inherited GetItem(Index));
end;

procedure TGalleryItems.SetItem(Index: Integer; Value: TGalleryItem);
begin
  inherited SetItem(Index, Value);
end;

function TGalleryItems.GetOwner: TPersistent;
begin
  Result := FGallery;
end;

function TGalleryItems.Insert(Index: Integer): TGalleryItem;
begin
  Result := AddItem(nil, Index);
end;

procedure TGalleryItems.Update(Item: TCollectionItem);
begin
  if (Item <> nil)
    then FGallery.UpdateItem(Item.Index)
    else FGallery.UpdateItems;
end;

{ TNiceGallery }

constructor TNiceGallery.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  Color := clWhite;
  Align := alTop;
  Height := 130;
  Delta := 0;
  MaxDelta := 0;
  HoverIndex := -1;
  LastHover := -1;
  LScrollerVisible := False;
  RScrollerVisible := False;
  MustScrollLeft := True;
  MustScrollRight := False;
  AutoScroll := True;
  CanScroll := False;

  FActive := False;
  FScrollerSize := 25;
  FScrollerArrowColor := clWhite;
  FScrollerColor := clBlack;
  FScrollerHoverArrowColor := clYellow;
  FScrollerHoverColor := clBlack;
  FHandPointCursor := True;
  FSelectionColor := clRed;
  FLargeChange := 30;
  FSmallChange := 3;
  FItems := TGalleryItems.Create(Self);
  FTimer := TTimer.Create(Self);
  with FTimer do
  begin
    Interval := 100;
    FTimer.OnTimer := TimerTimer;
    Enabled := False;
  end;
end;

destructor TNiceGallery.Destroy;
begin
  FTimer.Free;
  DeleteObject(RgnL);
  DeleteObject(RgnR);
  FItems.Free;
  inherited Destroy;
end;

function TNiceGallery.CreateItem: TGalleryItem;
begin
  Result := TGalleryItem.Create(FItems);
end;

procedure TNiceGallery.UpdateItem(Index: Integer);
begin
  Invalidate;
end;

procedure TNiceGallery.UpdateItems;
begin
  Recalculate;
  Invalidate;
end;

procedure TNiceGallery.Recalculate;
var
  x, n: Integer;
  h, w: Integer;
  mx: Integer;
  f: Integer;
  Rc: TRect;

begin

  FTimer.Enabled := False;
  n := 0;
  h := 0;
  for x := 0 to FItems.Count-1 do
  begin
    case (x mod 3) of
      0: h := Round(Height * 0.6);
      1: h := Round(Height * 0.85);
      2: h := Height;
    end;
    with FItems[x] do
    begin
      if not FBitmap.Empty
        then w := Round((FBitmap.Width / FBitmap.Height) * h)
        else w := 0;
      DisplayRect := Rect(n, 0, n + w, h);
    end;
    Inc(n, w);
  end;

  MaxDelta := Max(0, n - Width);
  CanScroll := MaxDelta > 0;

  LScroller := Rect(10, Height - FScrollerSize - 10, 10 + FScrollerSize, Height - 10);
  RScroller := Rect(Width - FScrollerSize - 10, Height - FScrollerSize - 10, Width - 10, Height - 10);

  with LScroller do
  begin
    f := FScrollerSize div 6;
    mx := Left + (3 * f);
    SetLength(LArrow, 7);
    LArrow[0] := Point(Left + f, Top + (3 * f));
    LArrow[1] := Point(mx, Top + f);
    LArrow[2] := Point(mx, Top + (2 * f));
    LArrow[3] := Point(Left + Round(4.5 * f), Top + (2 * f));
    LArrow[4] := Point(Left + Round(4.5 * f), Top + (4 * f));
    LArrow[5] := Point(mx, Top + (4 * f));
    LArrow[6] := Point(mx, Top + (5 * f));
  end;

  with RScroller do
  begin
    f := FScrollerSize div 6;
    mx := Left + (3 * f);
    SetLength(RArrow, 7);
    RArrow[0] := Point(Left + Round(1.5 * f), Top + (4 * f));
    RArrow[1] := Point(Left + Round(1.5 * f), Top + (2 * f));
    RArrow[2] := Point(mx, Top + (2 * f));
    RArrow[3] := Point(mx, Top + f);
    RArrow[4] := Point(Left + (5 * f), Top + (3 * f));
    RArrow[5] := Point(mx, Top + (5 * f));
    RArrow[6] := Point(mx, Top + (4 * f));
  end;

  DeleteObject(RgnL);
  DeleteObject(RgnR);

  Rc := LScroller;
  Rc.Right := Rc.Right + 1;
  Rc.Bottom := Rc.Bottom + 1;
  RgnL := CreateEllipticRgnIndirect(Rc);

  Rc := RScroller;
  Rc.Right := Rc.Right + 1;
  Rc.Bottom := Rc.Bottom + 1;
  RgnR := CreateEllipticRgnIndirect(Rc);

  CheckScroller;

  if not (csDesigning in ComponentState)
    then FTimer.Enabled := FActive;

end;

procedure TNiceGallery.CheckScroller;
begin
  LScrollerVisible := Delta > 0;
  RScrollerVisible := Delta < MaxDelta;
end;

procedure TNiceGallery.DrawItem(Index: Integer);
var
  Rgn: HRGN;
  Dummy, Rc, Rc2: TRect;
  LVis, RVis: Boolean;
begin
  case Index of

    -1:
      Exit;

    SCLEFTINDEX:
      if LScrollerVisible then
      begin
        with Canvas do
        begin
          Brush.Style := bsSolid;
          if (HoverIndex = Index) then
          begin
            Brush.Color := FScrollerHoverColor;
            Pen.Color := FScrollerHoverColor;
          end else
          begin
            Brush.Color := FScrollerColor;
            Pen.Color := FScrollerColor;
          end;
          Ellipse(LScroller);
          if (HoverIndex = Index) then
          begin
            Brush.Color := FScrollerHoverArrowColor;
            Pen.Color := FScrollerHoverArrowColor;
          end else
          begin
            Brush.Color := FScrollerArrowColor;
            Pen.Color := FScrollerArrowColor;
          end;
          Polygon(LArrow);
        end;
      end;

    SCRIGHTINDEX:
      if RScrollerVisible then
      begin
        with Canvas do
        begin
          Brush.Style := bsSolid;
          if (HoverIndex = Index) then
          begin
            Brush.Color := FScrollerHoverColor;
            Pen.Color := FScrollerHoverColor;
          end else
          begin
            Brush.Color := FScrollerColor;
            Pen.Color := FScrollerColor;
          end;
          Ellipse(RScroller);
          if (HoverIndex = Index) then
          begin
            Brush.Color := FScrollerHoverArrowColor;
            Pen.Color := FScrollerHoverArrowColor;
          end else
          begin
            Brush.Color := FScrollerArrowColor;
            Pen.Color := FScrollerArrowColor;
          end;
          Polygon(RArrow);
        end;
      end;

    else

      begin
        Rc := FItems[Index].DisplayRect;
        OffsetRect(Rc, -Delta, 0);
        if not IntersectRect(Dummy, Rc, ClientRect)
          then Exit;
        Rc2 := Rc;
        Rc2.Bottom := Height;
        LVis := LScrollerVisible and IntersectRect(Dummy, Rc2, LScroller);
        RVis := RScrollerVisible and IntersectRect(Dummy, Rc2, RScroller);
        Rgn := CreateRectRgnIndirect(Rc2);
        if LVis
          then CombineRgn(Rgn, Rgn, RgnL, RGN_DIFF);
        if RVis
          then CombineRgn(Rgn, Rgn, RgnR, RGN_DIFF);
        SelectClipRgn(Canvas.Handle, Rgn);
        with Canvas do
        begin
          Rc2 := Rc;
          Brush.Style := bsClear;
          Pen.Color := Color;
          Rectangle(Rc2);
          Rc2.Top := Rc.Bottom;
          Rc2.Bottom := Height;
          Brush.Style := bsSolid;
          Brush.Color := Color;
          FillRect(Rc2);
          if (Index = HoverIndex) then
          begin
            Brush.Style := bsClear;
            Pen.Color := clRed;
            Rectangle(Rc);
          end;
          InflateRect(Rc, -1, -1);
          StretchDraw(Rc, FItems[Index].FBitmap);
        end;
        DeleteObject(Rgn);
        SelectClipRgn(Canvas.Handle, 0);
      end;

  end;
end;

procedure TNiceGallery.Paint;
var
  x: Integer;
  n: Integer;
  Rc: TRect;
begin
  n := 0;
  if (FItems.Count > 0) then
  begin
    for x := 0 to FItems.Count-1
      do DrawItem(x);
    n := FItems[FItems.Count-1].DisplayRect.Right;
  end;
  if (n < Width) then
  begin
    Rc := Rect(n, 0, Width, Height);
    with Canvas do
    begin
      Brush.Style := bsSolid;
      Brush.Color := Color;
      FillRect(Rc);
    end;  
  end;
  if LScrollerVisible
    then DrawItem(SCLEFTINDEX);
  if RScrollerVisible
    then DrawItem(SCRIGHTINDEX);
end;

function TNiceGallery.GetItemAtXY(X, Y: Integer): Integer;
var
  P: TPoint;
  i: Integer;
begin
  Result := -1;
  p := Point(X, Y);
  if LScrollerVisible and PtInRect(LScroller, p)
    then Result := SCLEFTINDEX else
  if RScrollerVisible and PtInRect(RScroller, p)
    then Result := SCRIGHTINDEX
    else
    begin
      p := Point(X + Delta, Y);
      if (FItems.Count > 0) then
      begin
        for i := 0 to FItems.Count-1 do
        begin
          if PtInRect(FItems[i].DisplayRect, p) then
          begin
            Result := i;
            Break;
          end;
        end;
      end;
    end;
end;

procedure TNiceGallery.TimerTimer(Sender: TObject);
begin
  if (FItems.Count = 0)
    then Exit;
  if not CanScroll
    then Exit;  
  if MustScrollLeft then
  begin
    LastScroll := SCLEFTINDEX;
    Delta := Max(0, Delta - FSmallChange);
    CheckScroller;
    Invalidate;
    if (Delta = 0) then
    begin
      MustScrollLeft := False;
      if AutoScroll then
      begin
        LastScroll := SCRIGHTINDEX;
        MustScrollRight := True;
      end;
    end;
  end else
  if MustScrollRight then
  begin
    LastScroll := SCRIGHTINDEX;
    Delta := Min(MaxDelta, Delta + FSmallChange);
    CheckScroller;
    Invalidate;
    if (Delta = MaxDelta) then
    begin
      MustScrollRight := False;
      if AutoScroll then
      begin
        LastScroll := SCLEFTINDEX;
        MustScrollLeft := True;
      end;
    end;
  end;
end;

procedure TNiceGallery.MouseDown(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
var
  i: Integer;  
begin
  i := GetItemAtXY(X, Y);
  case i of
    SCLEFTINDEX:
      begin
        Delta := Max(0, Delta - FLargeChange);
        CheckScroller;
        Invalidate;
      end;
    SCRIGHTINDEX:
      begin
        Delta := Min(MaxDelta, Delta + FLargeChange);
        CheckScroller;
        Invalidate;
      end;
    else
      begin
        if Assigned(FOnClick)
          then FOnClick(Self, i);
      end;
  end;
  inherited;
end;

procedure TNiceGallery.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  i: Integer;
  IsItem: Boolean;
begin
  i := GetItemAtXY(X, Y);
  AutoScroll := False;
  if ((i = SCLEFTINDEX) or (i = SCRIGHTINDEX))
    then LastScroll := i;
  MustScrollLeft := i = SCLEFTINDEX;
  MustScrollRight := i = SCRIGHTINDEX;
  IsItem := (i <> -1) and (i <> SCLEFTINDEX) and (i <> SCRIGHTINDEX);
  if FHandPointCursor then
  begin
    case i of
      SCLEFTINDEX, SCRIGHTINDEX, -1:
        Cursor := crDefault;
      else
        Cursor := crHandPoint;
    end;
  end;
  if (i <> HoverIndex) then
  begin
    HoverIndex := i;
    if (LastHover >= 0) and (LastHover < FItems.Count)
      then DrawItem(LastHover);
    DrawItem(i);
    LastHover := i;
    if Assigned(FOnHover) then
    begin
      if IsItem
        then FOnHover(Self, i)
        else FOnHover(Self, -1);
    end;
    if ShowHint then
    begin
      if IsItem then
      begin
        Hint := FItems[i].FHint;
        Application.ActivateHint(Mouse.CursorPos);
      end else
        Application.HideHint;  
    end;
  end;
end;

procedure TNiceGallery.SetItems(Value: TGalleryItems);
begin
  Delta := 0;
  FItems.Assign(Value);
end;

procedure TNiceGallery.CMColorChanged(var Msg: TLMessage);
begin
  Invalidate;
end;

procedure TNiceGallery.WMSize(var Msg: TLMSize);
begin
  Delta := 0;
  Recalculate;
  Invalidate;
end;

procedure TNiceGallery.WMEraseBkgnd(var Msg: TLMessage);
begin
  Msg.Result := 1;
end;

procedure TNiceGallery.CMMouseLeave(var Msg: TLMessage);
begin
  HoverIndex := -1;
  if (LastHover >= 0) and (LastHover < FItems.Count)
    then DrawItem(LastHover);
  LastHover := -1;
  if FHandPointCursor
    then Cursor := crDefault;
  AutoScroll := True;
  if (LastScroll = SCLEFTINDEX)
    then MustScrollLeft := True else
  if (LastScroll = SCRIGHTINDEX)
    then MustScrollRight := True;
  inherited;
end;

procedure TNiceGallery.SetScrollerSize(Value: Integer);
begin
  if (FScrollerSize <> Value) then
  begin
    FScrollerSize := Value;
    Recalculate;
    Invalidate;
  end;
end;

procedure TNiceGallery.SetScrollerColor(Value: TColor);
begin
  if (FScrollerColor <> Value) then
  begin
    FScrollerColor := Value;
    Invalidate;
  end;
end;

procedure TNiceGallery.SetScrollerArrowColor(Value: TColor);
begin
  if (FScrollerArrowColor <> Value) then
  begin
    FScrollerArrowColor := Value;
    Invalidate;
  end;
end;

procedure TNiceGallery.SetActive(const Value: Boolean);
begin
  if (FActive <> Value) then
  begin
    FActive := Value;
    if not (csDesigning in ComponentState)
      then FTimer.Enabled := Value;
    if not Value
      then Delta := 0;
  end;
end;

end.
