{-------------------------------------------------------------------------------

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in compliance
with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS" basis,
WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
the specific language governing rights and limitations under the License.

     The Original Code is SortListViews.pas released at May 26th, 2007.
     The Initial Developer of the Original Code is Priyatna.
     (Website: http://www.priyatna.org Email: me@priyatna.org)
     All Rights Reserved.

     Contributors:
     - Bruce Christensen (twsocket.shaw.ca)

Alternatively, the contents of this file may be used under the terms of the
GNU General Public License Version 2 or later (the "GPL"), in which case
the provisions of the GPL are applicable instead of those above.
If you wish to allow use of your version of this file only under the terms
of the GPL and not to allow others to use your version of this file
under the MPL, indicate your decision by deleting the provisions above and
replace them with the notice and other provisions required by the GPL.
If you do not delete the provisions above, a recipient may use your version
of this file under either the MPL or the GPL.

-------------------------------------------------------------------------------}

unit SortListViews;

{$MODE Delphi}

interface

uses
  LCLIntf, LCLType, Controls, ComCtrls, Classes, Graphics, ImgList, SysUtils;

type
  TSLVOnCustomSort = procedure (Sender: TObject; ColIndex: Integer;
    Str1, Str2: string; var Res: Integer; var Handled: Boolean) of object;

  TSortListView = class(TListView)
  private
    UpdatingImages: Boolean;
    IndexUp, IndexDown: Integer;
    LastSortIndex: Integer;
    LastSortAsc: Boolean;
    FArrows: TImageList;
    FOnCustomSort: TSLVOnCustomSort;
    procedure AppendToImageList;
    procedure CreateImageList;
    procedure CreateArrows(Img: TCustomImageList);
    procedure ColumnClick(Sender: TObject; Column: TListColumn);
  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property OnCustomSort: TSLVOnCustomSort read FOnCustomSort write FOnCustomSort;
  end;

  procedure Register;


implementation

{ $R SortListViews.res}

procedure Register;
begin
  RegisterComponents('priyatna.org', [TSortListView]);
end;

{ TSortListView }

constructor TSortListView.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  ViewStyle := vsReport;
  RowSelect := True;
  HideSelection := False;
  ReadOnly := True;
  IndexUp := -1;
  IndexDown := -1;
  LastSortIndex := -1;
  LastSortAsc := True;
  OnColumnClick := ColumnClick;
end;

destructor TSortListView.Destroy;
begin
  UpdatingImages := True;
  if Assigned(FArrows)
    then FArrows.Free;
  inherited Destroy;
end;

procedure TSortListView.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if not (csDesigning in ComponentState) then
  begin
    if (AComponent = SmallImages) then
    begin
      if (Operation = opInsert)
        then AppendToImageList
        else CreateImageList;
    end;
  end;
end;

procedure TSortListView.Loaded;
begin
  inherited Loaded;
  if not (csDesigning in ComponentState) then
  begin
    if Assigned(SmallImages)
      then AppendToImageList
      else CreateImageList;
  end;
end;

procedure TSortListView.CreateArrows(Img: TCustomImageList);
var
  Bmp, Tmp: TBitmap;
begin
  if Assigned(Img) then
  begin
    Bmp := TBitmap.Create;
    Tmp := TBitmap.Create;
    try
      Tmp.Width := Img.Width;
      Tmp.Height := Img.Height;
      with Tmp.Canvas do
      begin
        Brush.Style := bsSolid;
        Brush.Color := clFuchsia;
        FillRect(Rect(0, 0, Tmp.Width, Tmp.Height));
      end;
      Img.AddMasked(Tmp, clFuchsia);

      Bmp.LoadFromResourceName(HInstance, 'SORT_UP');
      Bmp.Transparent := True;
      with Tmp.Canvas do
      begin
        Brush.Style := bsSolid;
        Brush.Color := clFuchsia;
        FillRect(Rect(0, 0, Tmp.Width, Tmp.Height));
      end;
      Tmp.Canvas.Draw((Tmp.Width - Bmp.Width) div 2, (Tmp.Height - Bmp.Height) div 2, Bmp);
      IndexUp := Img.AddMasked(Tmp, clFuchsia);

      Bmp.LoadFromResourceName(HInstance, 'SORT_DOWN');
      Bmp.Transparent := True;
      with Tmp.Canvas do
      begin
        Brush.Style := bsSolid;
        Brush.Color := clFuchsia;
        FillRect(Rect(0, 0, Tmp.Width, Tmp.Height));
      end;
      Tmp.Canvas.Draw((Tmp.Width - Bmp.Width) div 2, (Tmp.Height - Bmp.Height) div 2, Bmp);
      IndexDown := Img.AddMasked(Tmp, clFuchsia);
    finally
      Bmp.Free;
      Tmp.Free;
    end;
  end;
end;

procedure TSortListView.AppendToImageList;
begin
  if UpdatingImages
    then Exit;
  if Assigned(FArrows)
    then FArrows.Free;
  FArrows := nil;
  CreateArrows(SmallImages);
end;

procedure TSortListView.CreateImageList;
begin
  UpdatingImages := True;
  FArrows := TImageList.Create(Self);
  FArrows.Width := 16;
  FArrows.Height := 16;
  CreateArrows(FArrows);
  SmallImages := FArrows;
  UpdatingImages := False;
end;

function CustomSortProc(Item1, Item2: TListItem; ParamSort: Integer): Integer; stdcall;
var
  Index: Integer;
begin
  Index := TSortListView(ParamSort).LastSortIndex;
  if (Index = 0) then
  begin
    if TSortListView(ParamSort).LastSortAsc
      then Result := CompareText(TListItem(Item1).Caption, TListItem(Item2).Caption)
      else Result := -CompareText(TListItem(Item1).Caption, TListItem(Item2).Caption);
  end else
  begin
    if TSortListView(ParamSort).LastSortAsc
      then Result := CompareText(TListItem(Item1).SubItems[Index-1], TListItem(Item2).SubItems[Index-1])
      else Result := -CompareText(TListItem(Item1).SubItems[Index-1], TListItem(Item2).SubItems[Index-1]);
  end;
end;

function CustomSortProcEx(Item1, Item2: TListItem; ParamSort: Integer): Integer; stdcall;
var
  List: TSortListView;
  Index: Integer;
  Str1, Str2: string;
  Res: Integer;
  Handled: Boolean;
begin
  List := TSortListView(ParamSort);
  Index := List.LastSortIndex;
  Res := 0;
  Handled := False;
  if (Index = 0) then
  begin
    Str1 := TListItem(Item1).Caption;
    Str2 := TListItem(Item2).Caption;
  end else
  begin
    Str1 := TListItem(Item1).SubItems[Index-1];
    Str2 := TListItem(Item2).SubItems[Index-1];
  end;
  List.FOnCustomSort(List, Index, Str1, Str2, Res, Handled);
  if not Handled
    then Res := CompareText(Str1, Str2);
  if List.LastSortAsc
    then Result := Res
    else Result := -Res;
end;

procedure TSortListView.ColumnClick(Sender: TObject; Column: TListColumn);
begin
  if (LastSortIndex <> -1) and (LastSortIndex <> Column.Index)
    then Columns[LastSortIndex].ImageIndex := -1;
  if (LastSortIndex = Column.Index)
    then LastSortAsc := not LastSortAsc
    else LastSortAsc := True;
  LastSortIndex := Column.Index;
  if LastSortAsc
    then Column.ImageIndex := IndexUp
    else Column.ImageIndex := IndexDown;
  //if Assigned(FOnCustomSort)
  //  then CustomSort(@CustomSortProcEx, Integer(Self))
  //  else CustomSort(@CustomSortProc, Integer(Self));
end;

end.
